"""Python functions to easily generate a single-link URDF from a mesh file."""

from airo_models.urdf import dict_to_xml_str, material_dict, single_link_urdf_dict, write_urdf_to_tempfile


def mesh_geometry_dict(mesh_file: str) -> dict:
    geometry_dict = {"mesh": {"@filename": mesh_file}}
    return geometry_dict


def mesh_dict(mesh_file: str, name: str = "mesh") -> dict:
    geometry_dict = mesh_geometry_dict(mesh_file)
    mat_dict = material_dict((1.0, 1.0, 1.0, 1.0))
    mesh_dict_ = single_link_urdf_dict(name, geometry_dict, mat_dict)
    return mesh_dict_


def mesh_urdf(mesh_file: str, name: str = "mesh") -> str:
    mesh_dict_ = mesh_dict(mesh_file, name)
    urdf_str = dict_to_xml_str(mesh_dict_)
    return urdf_str


def mesh_urdf_path(mesh_file: str, name: str = "mesh") -> str:
    urdf_path = write_urdf_to_tempfile(mesh_dict(mesh_file, name), prefix=f"{name}_")
    return urdf_path


if __name__ == "__main__":
    print(mesh_urdf("/path/to/your/mesh.obj"))



"""Python functions to easily generate parametrized URDF boxes."""

from airo_models.urdf import dict_to_xml_str, material_dict, single_link_urdf_dict, write_urdf_to_tempfile


def box_geometry_dict(size: tuple[float, float, float]) -> dict:
    geometry_dict = {"box": {"@size": " ".join(map(str, size))}}
    return geometry_dict


def box_dict(
    size: tuple[float, float, float], name: str = "box", rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)
) -> dict:
    geometry_dict = box_geometry_dict(size)
    mat_dict = material_dict(rgba)
    box_dict_ = single_link_urdf_dict(name, geometry_dict, mat_dict)
    return box_dict_


def box_urdf(
    size: tuple[float, float, float], name: str = "box", rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)
) -> str:
    box_dict_ = box_dict(size, name, rgba)
    urdf_str = dict_to_xml_str(box_dict_)
    return urdf_str


def box_urdf_path(
    size: tuple[float, float, float], name: str = "box", rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)
) -> str:
    urdf_path = write_urdf_to_tempfile(box_dict(size, name, rgba), prefix=f"{name}_")
    return urdf_path


if __name__ == "__main__":
    print(box_urdf((0.5, 1.0, 2.0)))



import airo_models


def check_key_in_urdf(dict_: dict, key: str):
    """Check if a key is in a URDF dictionary."""
    for k, v in dict_.items():
        if k == key:
            return True
        elif isinstance(v, dict):
            if check_key_in_urdf(v, key):
                return True
        elif isinstance(v, list):
            for d in v:
                if check_key_in_urdf(d, key):
                    return True
    return False


def test_delete_key():
    """Test keys are deleted correctly from URDF dictionaries."""
    delete_test_dict = {
        "a": "1",
        "b": {
            "c": "2",
            "d": [{"f": "3"}, {"c": "4"}],
        },
        "e": 5,
    }

    airo_models.urdf.delete_key(delete_test_dict, "c")

    assert check_key_in_urdf(delete_test_dict, "a")
    assert check_key_in_urdf(delete_test_dict, "b")
    assert not check_key_in_urdf(delete_test_dict, "c")
    assert check_key_in_urdf(delete_test_dict, "d")
    assert check_key_in_urdf(delete_test_dict, "e")
    assert check_key_in_urdf(delete_test_dict, "f")



"""Python functions to easily generate parametrized URDF cylinders."""

from airo_models.urdf import dict_to_xml_str, material_dict, single_link_urdf_dict, write_urdf_to_tempfile


def cylinder_geometry_dict(length: float, radius: float) -> dict:
    geometry_dict = {"cylinder": {"@length": str(length), "@radius": str(radius)}}
    return geometry_dict


def cylinder_dict(
    length: float,
    radius: float,
    name: str = "cylinder",
    rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
) -> dict:
    geometry_dict = cylinder_geometry_dict(length, radius)
    mat_dict = material_dict(rgba)
    cylinder_dict_ = single_link_urdf_dict(name, geometry_dict, mat_dict)
    return cylinder_dict_


def cylinder_urdf(
    length: float,
    radius: float,
    name: str = "cylinder",
    rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
) -> str:
    cylinder_dict_ = cylinder_dict(length, radius, name, rgba)
    urdf_str = dict_to_xml_str(cylinder_dict_)
    return urdf_str


def cylinder_urdf_path(
    length: float,
    radius: float,
    name: str = "cylinder",
    rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
) -> str:
    urdf_path = write_urdf_to_tempfile(cylinder_dict(length, radius, name, rgba), prefix=f"{name}_")
    return urdf_path


if __name__ == "__main__":
    print(cylinder_urdf(2.0, 0.5))



# init files are still required in modern python!
# https://peps.python.org/pep-0420/ introduced implicit namespace packages
# but for building and many toolings, you still need to have __init__ files (at least in the root of the package).
# e.g. if you remove this init file and try to build with pip install .
# you won't be able to import the dummy module.
import airo_models.primitives as primitives  # noqa: F401
import airo_models.urdf as urdf  # noqa: F401
from airo_models.primitives.box import *  # noqa: F401, F403
from airo_models.primitives.cylinder import *  # noqa: F401, F403
from airo_models.primitives.mesh import *  # noqa: F401, F403
from airo_models.primitives.sphere import *  # noqa: F401, F403

# Not sure if this relative import is needed. Should investigate.
from .files import AIRO_MODEL_NAMES, get_urdf_path

__all__ = ["urdf", "get_urdf_path", "AIRO_MODEL_NAMES"]



"""Reading, writing and recurvise operations for URDFs, such as deleting keys or replacing values."""

import os
import tempfile

import xmltodict


def read_urdf(urdf_path: str) -> dict:
    """Reads a URDF (xml) file into a Python dictionary.
    Childern elements can be accessed by using the tag name as key e.g. urdf["robot"].
    Attributes can be accessed by using the "@" prefix e.g. link["@name"].

    Args:
        urdf_path: path to the URDF file.

    Returns:
        The URDF as a dictionary.
    """
    with open(urdf_path, "r") as file:
        xml_content = file.read()
    urdf = xmltodict.parse(xml_content)
    return urdf


def write_urdf_to_file(dict_: dict, urdf_path: str) -> None:
    urdf_str = dict_to_xml_str(dict_)
    with open(urdf_path, "w") as f:
        f.write(urdf_str)


def write_urdf_to_tempfile(dict_: dict, original_urdf_path: str | None = None, prefix: str = "urdf_model_") -> str:
    if original_urdf_path is not None:
        make_paths_absolute(dict_, original_urdf_path)
    urdf_str = dict_to_xml_str(dict_)
    urdf_file = tempfile.NamedTemporaryFile(prefix=prefix, suffix=".urdf", delete=False)
    urdf_file.write(urdf_str.encode())
    urdf_path = urdf_file.name
    return urdf_path


def dict_to_xml_str(dict_: dict) -> str:
    if not dict_:
        return ""
    return xmltodict.unparse(dict_, pretty=True, indent="  ", short_empty_elements=True)


def delete_key(dict_: dict, key: str) -> None:
    for k, v in list(dict_.items()):
        if k == key:
            del dict_[k]
        # Recurse
        elif isinstance(v, dict):
            delete_key(v, key)
        elif isinstance(v, list):
            for i in v:
                delete_key(i, key)


def replace_value(dict_: dict, key: str, value: str, new_value: str) -> None:
    for k, v in dict_.items():
        if key == key and v == value:
            dict_[k] = new_value
        # Recurse
        elif isinstance(v, dict):
            replace_value(v, key, value, new_value)
        elif isinstance(v, list):
            for d in v:
                replace_value(d, key, value, new_value)


def make_path_absolute(rel_path: str, urdf_path: str) -> str:
    return os.path.normpath(os.path.join(os.path.abspath(os.path.dirname(urdf_path)), rel_path))


def make_paths_absolute(dict_: dict, urdf_path: str) -> None:
    for key, value in dict_.items():
        if key == "@filename" and not os.path.isabs(value):
            dict_[key] = make_path_absolute(value, urdf_path)
        # Recurse
        elif isinstance(value, dict):
            make_paths_absolute(value, urdf_path)
        elif isinstance(value, list):
            for item in value:
                make_paths_absolute(item, urdf_path)


def single_link_urdf_dict(name: str, geometry_dict: dict, material_dict: dict) -> dict:
    """Generate the basic structure of URDF for a single link with a given geometry.

    Corresponds to this URDF:
    ```xml
    <robot name="name">
        <link name="base_link">
            <visual>
                <geometry>
                    ...
                </geometry>
            </visual>
            <collision>
                <geometry>
                    ...
                </geometry>
            </collision>
        </link>
    </robot>
    ```

    Used by the functions that generate URDF primitives.
    """
    return {
        "robot": {
            "@name": name,
            "link": {
                "@name": "base_link",
                "visual": {"geometry": geometry_dict, "material": material_dict},
                "collision": {"geometry": geometry_dict},
            },
        }
    }


def make_static(dict_: dict) -> None:
    """Make all joints in the URDF fixed and remove mimic and transmission keys."""

    joint_types = ["revolute", "continuous", "prismatic", "floating", "planar"]
    for joint_type in joint_types:
        replace_value(dict_, "@type", joint_type, "fixed")

    delete_key(dict_, "mimic")
    delete_key(dict_, "transmission")


def _get_robot_element_by_name(urdf_dict: dict, element_type: str, element_name: str) -> dict | None:
    """Searches for a URDF link or joint with the specified name.
    Args:
        urdf_dict: A dictionary containing the parsed URDF data.
        element_type: The type of element to search for ('link' or 'joint').
        element_name: The name of the element to find.

    Returns:
        The URDF element dictionary if found, otherwise None.
    """
    if element_type not in ["link", "joint"]:
        raise ValueError("Invalid element_type. Must be 'link' or 'joint'")

    for element in urdf_dict["robot"][element_type]:
        if element["@name"] == element_name:
            return element
    return None  # Element not found


def get_link_by_name(urdf_dict: dict, link_name: str) -> dict | None:
    """Searches for a URDF link with the specified name.

    Args:
        urdf_dict: A dictionary containing the parsed URDF data.
        link_name: The name of the link to find.

    Returns:
        The URDF link dictionary if found, otherwise None.
    """
    return _get_robot_element_by_name(urdf_dict, "link", link_name)


def get_joint_by_name(urdf_dict: dict, joint_name: str) -> dict | None:
    """Searches for a URDF joint with the specified name.

    Args:
        urdf_dict: A dictionary containing the parsed URDF data.
        joint_name: The name of the joint to find.

    Returns:
        The URDF joint dictionary if found, otherwise None.
    """
    return _get_robot_element_by_name(urdf_dict, "joint", joint_name)


def material_dict(rgba: tuple[float, float, float, float]) -> dict:
    material_dict = {"color": {"@rgba": " ".join(map(str, rgba))}}
    return material_dict



import os

AIRO_MODEL_NAMES = ["ur3e", "ur5e", "robotiq_2f_85"]


def get_urdf_path(name: str) -> str:
    """Get the absolute path to a URDF file using a name e.g. "ur5e".

    Args:
        name: the name of robot or gripper you want the URDF file for

    Returns:
        urdf_path: the absolute path to the URDF file
    """

    airo_models_dir = os.path.dirname(__file__)

    name_to_urdf_path = {
        "ur3e": os.path.join(airo_models_dir, "arms", "ur3e", "ur3e.urdf"),
        "ur5e": os.path.join(airo_models_dir, "arms", "ur5e", "ur5e.urdf"),
        "robotiq_2f_85": os.path.join(airo_models_dir, "grippers", "robotiq_2f_85", "urdf", "robotiq_2f_85.urdf"),
        "schunk_egk40": os.path.join(airo_models_dir, "grippers", "schunk_egk40", "urdf", "schunk_egk40.urdf"),
        "schunk_egk40_magneto": os.path.join(
            airo_models_dir, "grippers", "schunk_egk40_magneto", "schunk_egk40_magneto.urdf"
        ),
        "kelo_robile": os.path.join(airo_models_dir, "mobile_platforms", "kelo_robile", "urdf", "mobi.urdf"),
        "kelo_robile_battery": os.path.join(
            airo_models_dir, "mobile_platforms", "kelo_robile", "urdf", "battery_brick.urdf"
        ),
        "kelo_robile_cpu": os.path.join(airo_models_dir, "mobile_platforms", "kelo_robile", "urdf", "cpu_brick.urdf"),
        "kelo_robile_wheel": os.path.join(
            airo_models_dir, "mobile_platforms", "kelo_robile", "urdf", "wheel_brick.urdf"
        ),
        "zed2i": os.path.join(airo_models_dir, "cameras", "zed", "zed2i.urdf"),
        "zedm": os.path.join(airo_models_dir, "cameras", "zed", "zedm.urdf"),
        "d435": os.path.join(airo_models_dir, "cameras", "realsense", "d435", "d435.urdf"),
        "table8080": os.path.join(airo_models_dir, "environment", "tables", "table8080.urdf"),
        "mounting_plate_ur3e": os.path.join(
            airo_models_dir, "environment", "mounting_plates", "mounting_plate_ur3e.urdf"
        ),
        "mounting_plate_ur5e": os.path.join(
            airo_models_dir, "environment", "mounting_plates", "mounting_plate_ur5e.urdf"
        ),
    }

    if name in name_to_urdf_path:
        urdf_path = os.path.abspath(name_to_urdf_path[name])
        return urdf_path
    else:
        raise ValueError(f"We have no URDF for name: {name}")


if __name__ == "__main__":
    for name in AIRO_MODEL_NAMES:
        print(f"{name}: {get_urdf_path(name)}")



"""Python functions to easily generate parametrized URDF spheres."""

from airo_models.urdf import dict_to_xml_str, material_dict, single_link_urdf_dict, write_urdf_to_tempfile


def sphere_geometry_dict(radius: float) -> dict:
    geometry_dict = {"sphere": {"@radius": str(radius)}}
    return geometry_dict


def sphere_dict(
    radius: float, name: str = "sphere", rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)
) -> dict:
    geometry_dict = sphere_geometry_dict(radius)
    mat_dict = material_dict(rgba)
    sphere_dict_ = single_link_urdf_dict(name, geometry_dict, mat_dict)
    return sphere_dict_


def sphere_urdf(
    radius: float, name: str = "sphere", rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)
) -> str:
    sphere_dict_ = sphere_dict(radius, name, rgba)
    urdf_str = dict_to_xml_str(sphere_dict_)
    return urdf_str


def sphere_urdf_path(
    radius: float, name: str = "sphere", rgba: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)
) -> str:
    urdf_path = write_urdf_to_tempfile(sphere_dict(radius, name, rgba), prefix=f"{name}_")
    return urdf_path


if __name__ == "__main__":
    print(sphere_urdf(0.5))



import os

import airo_models


def test_existence():
    """Test whether the URDF files were included in the pip install.

    Try this test with both:
    pip install -e .
    pip install .
    """
    for name in airo_models.AIRO_MODEL_NAMES:
        urdf_path = airo_models.get_urdf_path(name)
        assert os.path.exists(urdf_path), f"URDF file for {name} does not exist at {urdf_path}"

    # Also test long-form import
    for name in airo_models.files.AIRO_MODEL_NAMES:
        urdf_path = airo_models.files.get_urdf_path(name)
        assert os.path.exists(urdf_path), f"URDF file for {name} does not exist at {urdf_path}"
